@@ -2,13 +2,13 @@ from __future__ import absolute_import, division, print_function
 
 from os.path import join
 
-from .common import name_prefix, error_and_exit
-from ..config import envs_dirs, root_dir
-
+from .common import name_prefix
+from ..base.context import context
+from ..exceptions import CondaIOError
 
 def read_message(fn):
     res = []
-    for envs_dir in envs_dirs:
+    for envs_dir in context.envs_dirs:
         path = join(envs_dir, '.conda-help', fn)
         try:
             with open(path) as fi:
@@ -36,8 +36,8 @@ Missing write permissions in: ${root_dir}
 #
 # $ conda create -n my_${name} --clone=${prefix}
 """
-    msg = msg.replace('${root_dir}', root_dir)
+    msg = msg.replace('${root_dir}', context.root_dir)
     msg = msg.replace('${prefix}', prefix)
     msg = msg.replace('${name}', name_prefix(prefix))
     msg = msg.replace('${command}', command)
-    error_and_exit(msg, json=json, error_type='RootNotWritable')
+    raise CondaIOError(msg)
