@@ -4,24 +4,33 @@
 # conda is distributed under the terms of the BSD 3-clause license.
 # Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.
 
-from __future__ import print_function, division, absolute_import
+from __future__ import absolute_import, division, print_function
 
-import base64
 import cgi
 import email
 import ftplib
 import mimetypes
 import os
 import platform
-import re
 import requests
 import tempfile
+from base64 import b64decode
 from io import BytesIO
 from logging import getLogger
+from requests.adapters import HTTPAdapter
+from requests.auth import AuthBase, _basic_auth_str
+from requests.cookies import extract_cookies_to_jar
+from requests.utils import get_auth_from_url, get_netrc_auth
 
 from . import __version__ as VERSION
-from .compat import urlparse, StringIO
-from .config import platform as config_platform, ssl_verify, get_proxy_servers
+from ._vendor.auxlib.ish import dals
+from .base.context import context
+from .common.disk import rm_rf
+from .common.url import (add_username_and_password, get_proxy_username_and_pass,
+                         split_anaconda_token, url_to_path, url_to_s3_info, urlparse)
+from .compat import StringIO, iteritems
+from .exceptions import AuthenticationError, ProxyError
+from .gateways.anaconda_client import read_binstar_tokens
 from .utils import gnu_get_libc_version
 
 RETRIES = 3
@@ -36,10 +45,10 @@ _user_agent = ("conda/{conda_ver} "
                "{system}/{kernel} {dist}/{ver}")
 
 glibc_ver = gnu_get_libc_version()
-if config_platform == 'linux':
+if context.platform == 'linux':
     distinfo = platform.linux_distribution()
     dist, ver = distinfo[0], distinfo[1]
-elif config_platform == 'osx':
+elif context.platform == 'osx':
     dist = 'OSX'
     ver = platform.mac_ver()[0]
 else:
@@ -55,28 +64,6 @@ user_agent = _user_agent.format(conda_ver=VERSION,
 if glibc_ver:
     user_agent += " glibc/{}".format(glibc_ver)
 
-# Modified from code in pip/download.py:
-
-# Copyright (c) 2008-2014 The pip developers (see AUTHORS.txt file)
-#
-# Permission is hereby granted, free of charge, to any person obtaining
-# a copy of this software and associated documentation files (the
-# "Software"), to deal in the Software without restriction, including
-# without limitation the rights to use, copy, modify, merge, publish,
-# distribute, sublicense, and/or sell copies of the Software, and to
-# permit persons to whom the Software is furnished to do so, subject to
-# the following conditions:
-#
-# The above copyright notice and this permission notice shall be
-# included in all copies or substantial portions of the Software.
-#
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
-# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
-# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
-# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 class CondaSession(requests.Session):
 
@@ -87,13 +74,15 @@ class CondaSession(requests.Session):
 
         super(CondaSession, self).__init__(*args, **kwargs)
 
-        proxies = get_proxy_servers()
+        self.auth = CondaHttpAuth()  # TODO: should this just be for certain protocol adapters?
+
+        proxies = context.proxy_servers
         if proxies:
             self.proxies = proxies
 
         # Configure retries
         if retries:
-            http_adapter = requests.adapters.HTTPAdapter(max_retries=retries)
+            http_adapter = HTTPAdapter(max_retries=retries)
             self.mount("http://", http_adapter)
             self.mount("https://", http_adapter)
 
@@ -108,7 +97,103 @@ class CondaSession(requests.Session):
 
         self.headers['User-Agent'] = user_agent
 
-        self.verify = ssl_verify
+        self.verify = context.ssl_verify
+
+        if context.client_tls_cert_key:
+            self.cert = (context.client_tls_cert, context.client_tls_cert_key)
+        elif context.client_tls_cert:
+            self.cert = context.client_tls_cert
+
+
+class CondaHttpAuth(AuthBase):
+    # TODO: make this class thread-safe by adding some of the requests.auth.HTTPDigestAuth() code
+
+    def __call__(self, request):
+        request.url = CondaHttpAuth.add_binstar_token(request.url)
+        self._apply_basic_auth(request)
+        request.register_hook('response', self.handle_407)
+        return request
+
+    @staticmethod
+    def _apply_basic_auth(request):
+        # this logic duplicated from Session.prepare_request and PreparedRequest.prepare_auth
+        url_auth = get_auth_from_url(request.url)
+        auth = url_auth if any(url_auth) else None
+
+        if auth is None:
+            # look for auth information in a .netrc file
+            auth = get_netrc_auth(request.url)
+
+        if isinstance(auth, tuple) and len(auth) == 2:
+            request.headers['Authorization'] = _basic_auth_str(*auth)
+
+        return request
+
+    @staticmethod
+    def add_binstar_token(url):
+        clean_url, token = split_anaconda_token(url)
+        if not token:
+            for binstar_url, token in iteritems(read_binstar_tokens()):
+                if clean_url.startswith(binstar_url):
+                    log.debug("Adding anaconda token for url <%s>", clean_url)
+                    from conda.models.channel import Channel
+                    channel = Channel(clean_url)
+                    channel.token = token
+                    return channel.url(with_credentials=True)
+        return url
+
+    def handle_407(self, response, **kwargs):
+        """
+        Prompts the user for the proxy username and password and modifies the
+        proxy in the session object to include it.
+
+        This method is modeled after
+          * requests.auth.HTTPDigestAuth.handle_401()
+          * requests.auth.HTTPProxyAuth
+          * the previous conda.fetch.handle_proxy_407()
+
+        It both adds 'username:password' to the proxy URL, as well as adding a
+        'Proxy-Authorization' header.  If any of this is incorrect, please file an issue.
+
+        """
+        # kwargs = {'verify': True, 'cert': None, 'proxies': OrderedDict(), 'stream': False,
+        #           'timeout': (3.05, 60)}
+
+        if response.status_code != 407:
+            return response
+
+        # Consume content and release the original connection
+        # to allow our new request to reuse the same one.
+        response.content
+        response.close()
+
+        proxies = kwargs.pop('proxies')
+
+        proxy_scheme = urlparse(response.url).scheme
+        if proxy_scheme not in proxies:
+            raise ProxyError(dals("""Could not find a proxy for %r. See
+            http://conda.pydata.org/docs/html#configure-conda-for-use-behind-a-proxy-server
+            for more information on how to configure proxies.""" % proxy_scheme))
+
+        # fix-up proxy_url with username & password
+        proxy_url = proxies[proxy_scheme]
+        username, password = get_proxy_username_and_pass(proxy_scheme)
+        proxy_url = add_username_and_password(proxy_url, username, password)
+        proxy_authorization_header = _basic_auth_str(username, password)
+        proxies[proxy_scheme] = proxy_url
+        kwargs['proxies'] = proxies
+
+        prep = response.request.copy()
+        extract_cookies_to_jar(prep._cookies, response.request, response.raw)
+        prep.prepare_cookies(prep._cookies)
+        prep.headers['Proxy-Authorization'] = proxy_authorization_header
+
+        _response = response.connection.send(prep, **kwargs)
+        _response.history.append(response)
+        _response.request = prep
+
+        return _response
+
 
 class S3Adapter(requests.adapters.BaseAdapter):
 
@@ -116,8 +201,7 @@ class S3Adapter(requests.adapters.BaseAdapter):
         super(S3Adapter, self).__init__()
         self._temp_file = None
 
-    def send(self, request, stream=None, timeout=None, verify=None, cert=None,
-             proxies=None):
+    def send(self, request, stream=None, timeout=None, verify=None, cert=None, proxies=None):
 
         resp = requests.models.Response()
         resp.status_code = 200
@@ -150,7 +234,7 @@ class S3Adapter(requests.adapters.BaseAdapter):
 
         conn = boto.connect_s3()
 
-        bucket_name, key_string = url_to_S3_info(request.url)
+        bucket_name, key_string = url_to_s3_info(request.url)
 
         # Get the bucket without validation that it exists and that we have
         # permissions to list its contents.
@@ -172,7 +256,7 @@ class S3Adapter(requests.adapters.BaseAdapter):
                 "Content-Type": content_type,
                 "Content-Length": key.size,
                 "Last-Modified": modified,
-                })
+            })
 
             _, self._temp_file = tempfile.mkstemp()
             key.get_contents_to_filename(self._temp_file)
@@ -186,24 +270,12 @@ class S3Adapter(requests.adapters.BaseAdapter):
 
     def close(self):
         if self._temp_file:
-            os.remove(self._temp_file)
-
-
-def url_to_S3_info(url):
-    """
-    Convert a S3 url to a tuple of bucket and key
-    """
-    parsed_url = requests.packages.urllib3.util.url.parse_url(url)
-    assert parsed_url.scheme == 's3', (
-        "You can only use s3: urls (not %r)" % url)
-    bucket, key = parsed_url.host, parsed_url.path
-    return bucket, key
+            rm_rf(self._temp_file)
 
 
 class LocalFSAdapter(requests.adapters.BaseAdapter):
 
-    def send(self, request, stream=None, timeout=None, verify=None, cert=None,
-             proxies=None):
+    def send(self, request, stream=None, timeout=None, verify=None, cert=None, proxies=None):
         pathname = url_to_path(request.url)
 
         resp = requests.models.Response()
@@ -233,24 +305,6 @@ class LocalFSAdapter(requests.adapters.BaseAdapter):
         pass
 
 
-def url_to_path(url):
-    """
-    Convert a file: URL to a path.
-    """
-    assert url.startswith('file:'), (
-        "You can only turn file: urls into filenames (not %r)" % url)
-    path = url[len('file:'):].lstrip('/')
-    path = urlparse.unquote(path)
-    if _url_drive_re.match(path):
-        path = path[0] + ':' + path[2:]
-    elif not path.startswith(r'\\'):
-        # if not a Windows UNC path
-        path = '/' + path
-    return path
-
-_url_drive_re = re.compile('^([a-z])[:|]', re.I)
-
-
 # Taken from requests-ftp
 # (https://github.com/Lukasa/requests-ftp/blob/master/requests_ftp/ftp.py)
 
@@ -293,6 +347,9 @@ class FTPAdapter(requests.adapters.BaseAdapter):
 
         # Sort out the timeout.
         timeout = kwargs.get('timeout', None)
+        if not isinstance(timeout, int):
+            # https://github.com/conda/conda/pull/3392
+            timeout = 10
 
         # Establish the connection and login if needed.
         self.conn = ftplib.FTP()
@@ -405,12 +462,12 @@ class FTPAdapter(requests.adapters.BaseAdapter):
             # second part. Check that we have the right kind of auth though.
             encoded_components = auth_header.split()[:2]
             if encoded_components[0] != 'Basic':
-                raise AuthError('Invalid form of Authentication used.')
+                raise AuthenticationError('Invalid form of Authentication used.')
             else:
                 encoded = encoded_components[1]
 
             # Decode the base64 encoded string.
-            decoded = base64.b64decode(encoded)
+            decoded = b64decode(encoded)
 
             # The string is of the form 'username:password'. Split on the
             # colon.
@@ -428,7 +485,7 @@ class FTPAdapter(requests.adapters.BaseAdapter):
         of urlparse's craziness.'''
         url = request.url
         # scheme, netloc, path, params, query, fragment = urlparse(url)
-        parsed = urlparse.urlparse(url)
+        parsed = urlparse(url)
         path = parsed.path
 
         # If there is a slash on the front of the path, chuck it.
@@ -452,18 +509,16 @@ def data_callback_factory(variable):
     return callback
 
 
-class AuthError(Exception):
-    '''Denotes an error with authentication.'''
-    pass
-
 def build_text_response(request, data, code):
     '''Build a response for textual data.'''
     return build_response(request, data, code, 'ascii')
 
+
 def build_binary_response(request, data, code):
     '''Build a response for data whose encoding is unknown.'''
     return build_response(request, data, code,  None)
 
+
 def build_response(request, data, code, encoding):
     '''Builds a response object from the data returned by ftplib, using the
     specified encoding.'''
@@ -484,6 +539,7 @@ def build_response(request, data, code, encoding):
     response = requests.hooks.dispatch_hook('response', request.hooks, response)
     return response
 
+
 def parse_multipart_files(request):
     '''Given a prepared reqest, return a file-like object containing the
     original data. This is pretty hacky.'''
@@ -506,51 +562,3 @@ def parse_multipart_files(request):
     buf.seek(0)
 
     return buf
-
-# Taken from urllib3 (actually
-# https://github.com/shazow/urllib3/pull/394). Once it is fully upstreamed to
-# requests.packages.urllib3 we can just use that.
-
-
-def unparse_url(U):
-    """
-    Convert a :class:`.Url` into a url
-
-    The input can be any iterable that gives ['scheme', 'auth', 'host',
-    'port', 'path', 'query', 'fragment']. Unused items should be None.
-
-    This function should more or less round-trip with :func:`.parse_url`. The
-    returned url may not be exactly the same as the url inputted to
-    :func:`.parse_url`, but it should be equivalent by the RFC (e.g., urls
-    with a blank port).
-
-
-    Example: ::
-
-        >>> Url = parse_url('http://google.com/mail/')
-        >>> unparse_url(Url)
-        'http://google.com/mail/'
-        >>> unparse_url(['http', 'username:password', 'host.com', 80,
-        ... '/path', 'query', 'fragment'])
-        'http://username:password@host.com:80/path?query#fragment'
-    """
-    scheme, auth, host, port, path, query, fragment = U
-    url = ''
-
-    # We use "is not None" we want things to happen with empty strings (or 0 port)
-    if scheme is not None:
-        url = scheme + '://'
-    if auth is not None:
-        url += auth + '@'
-    if host is not None:
-        url += host
-    if port is not None:
-        url += ':' + str(port)
-    if path is not None:
-        url += path
-    if query is not None:
-        url += '?' + query
-    if fragment is not None:
-        url += '#' + fragment
-
-    return url
