@@ -18,14 +18,8 @@ These API functions have argument names referring to:
                  be the "default" environment (i.e. sys.prefix),
                  but is otherwise something like '/opt/anaconda/envs/foo',
                  or even any prefix, e.g. '/home/joe/myenv'
-
-Also, this module is directly invoked by the (self extracting (sfx)) tarball
-installer to create the initial environment, therefore it needs to be
-standalone, i.e. not import any other parts of `conda` (only depend on
-the standard library).
 """
-
-from __future__ import print_function, division, absolute_import
+from __future__ import absolute_import, division, print_function, unicode_literals
 
 import errno
 import functools
@@ -36,65 +30,31 @@ import re
 import shlex
 import shutil
 import stat
+import struct
 import subprocess
 import sys
 import tarfile
-import tempfile
-import time
 import traceback
-from os.path import (abspath, basename, dirname, isdir, isfile, islink,
-                     join, normpath, normcase)
+from collections import namedtuple
+from enum import Enum
+from itertools import chain
+from os.path import (abspath, basename, dirname, exists, isdir, isfile, islink, join, normcase,
+                     normpath)
+
+from . import CondaError
+from .base.constants import UTF8
+from .base.context import context
+from .common.disk import exp_backoff_fn, rm_rf
+from .common.url import path_to_url
+from .exceptions import CondaOSError, LinkError, PaddingError
+from .lock import DirectoryLock, FileLock
+from .models.channel import Channel
+from .utils import on_win
 
 
-on_win = bool(sys.platform == "win32")
+# conda-build compatibility
+from .common.disk import delete_trash, move_to_trash, move_path_to_trash  # NOQA
 
-try:
-    from conda.lock import Locked
-    from conda.utils import win_path_to_unix, url_path
-    from conda.config import remove_binstar_tokens, pkgs_dirs, url_channel
-    import conda.config as config
-except ImportError:
-    # Make sure this still works as a standalone script for the Anaconda
-    # installer.
-    pkgs_dirs = [sys.prefix]
-
-    class Locked(object):
-        def __init__(self, *args, **kwargs):
-            pass
-
-        def __enter__(self):
-            pass
-
-        def __exit__(self, exc_type, exc_value, traceback):
-            pass
-
-    def win_path_to_unix(path, root_prefix=""):
-        """Convert a path or ;-separated string of paths into a unix representation
-
-        Does not add cygdrive.  If you need that, set root_prefix to "/cygdrive"
-        """
-        path_re = '(?<![:/^a-zA-Z])([a-zA-Z]:[\/\\\\]+(?:[^:*?"<>|]+[\/\\\\]+)*[^:*?"<>|;\/\\\\]+?(?![a-zA-Z]:))'  # noqa
-
-        def translation(found_path):
-            found = found_path.group(1).replace("\\", "/").replace(":", "")
-            return root_prefix + "/" + found
-        return re.sub(path_re, translation, path).replace(";/", ":/")
-
-    def url_path(path):
-        path = abspath(path)
-        if on_win:
-            path = '/' + path.replace(':', '|').replace('\\', '/')
-        return 'file://%s' % path
-
-    # There won't be any binstar tokens in the installer anyway
-    def remove_binstar_tokens(url):
-        return url
-
-    # A simpler version of url_channel will do
-    def url_channel(url):
-        return url.rsplit('/', 2)[0] + '/' if url and '/' in url else None, 'defaults'
-
-    pkgs_dirs = [join(sys.prefix, 'pkgs')]
 
 if on_win:
     import ctypes
@@ -115,14 +75,14 @@ if on_win:
     def win_hard_link(src, dst):
         "Equivalent to os.link, using the win32 CreateHardLink call."
         if not CreateHardLink(dst, src, None):
-            raise OSError('win32 hard link failed')
+            raise CondaOSError('win32 hard link failed')
 
     def win_soft_link(src, dst):
         "Equivalent to os.symlink, using the win32 CreateSymbolicLink call."
         if CreateSymbolicLink is None:
-            raise OSError('win32 soft link not supported')
+            raise CondaOSError('win32 soft link not supported')
         if not CreateSymbolicLink(dst, src, isdir(src)):
-            raise OSError('win32 soft link failed')
+            raise CondaOSError('win32 soft link failed')
 
     def win_conda_bat_redirect(src, dst, shell):
         """Special function for Windows XP where the `CreateSymbolicLink`
@@ -171,22 +131,17 @@ if on_win:
 log = logging.getLogger(__name__)
 stdoutlog = logging.getLogger('stdoutlog')
 
-class NullHandler(logging.Handler):
-    """ Copied from Python 2.7 to avoid getting
-        `No handlers could be found for logger "patch"`
-        http://bugs.python.org/issue16539
-    """
 
-    def handle(self, record):
-        pass
+SHEBANG_REGEX = re.compile(br'^(#!((?:\\ |[^ \n\r])+)(.*))')
 
-    def emit(self, record):
-        pass
 
-    def createLock(self):
-        self.lock = None
+class FileMode(Enum):
+    text = 'text'
+    binary = 'binary'
+
+    def __str__(self):
+        return "%s" % self.value
 
-log.addHandler(NullHandler())
 
 LINK_HARD = 1
 LINK_SOFT = 2
@@ -215,13 +170,14 @@ def _link(src, dst, linktype=LINK_HARD):
         else:
             shutil.copy2(src, dst)
     else:
-        raise Exception("Did not expect linktype=%r" % linktype)
+        raise CondaError("Did not expect linktype=%r" % linktype)
 
 
 def _remove_readonly(func, path, excinfo):
     os.chmod(path, stat.S_IWRITE)
     func(path)
 
+
 def warn_failed_remove(function, path, exc_info):
     if exc_info[1].errno == errno.EACCES:
         log.warn("Cannot remove, permission denied: {0}".format(path))
@@ -231,135 +187,68 @@ def warn_failed_remove(function, path, exc_info):
         log.warn("Cannot remove, unknown reason: {0}".format(path))
 
 
-def exp_backoff_fn(fn, *args):
-    """Mostly for retrying file operations that fail on Windows due to virus scanners"""
-    if not on_win:
-        return fn(*args)
+def yield_lines(path):
+    """Generator function for lines in file.  Empty generator if path does not exist.
 
-    import random
-    # with max_tries = 5, max total time ~= 3.2 sec
-    # with max_tries = 6, max total time ~= 6.5 sec
-    max_tries = 6
-    for n in range(max_tries):
-        try:
-            result = fn(*args)
-        except (OSError, IOError) as e:
-            log.debug(repr(e))
-            if e.errno in (errno.EPERM, errno.EACCES):
-                if n == max_tries-1:
-                    raise
-                time.sleep(((2 ** n) + random.random()) * 0.1)
-            else:
-                raise
-        else:
-            return result
+    Args:
+        path (str): path to file
 
+    Returns:
+        iterator: each line in file, not starting with '#'
 
-def rm_rf(path, max_retries=5, trash=True):
-    """
-    Completely delete path
-    max_retries is the number of times to retry on failure. The default is
-    5. This only applies to deleting a directory.
-    If removing path fails and trash is True, files will be moved to the trash directory.
-    """
-    if islink(path) or isfile(path):
-        # Note that we have to check if the destination is a link because
-        # exists('/path/to/dead-link') will return False, although
-        # islink('/path/to/dead-link') is True.
-        try:
-            os.unlink(path)
-            return
-        except (OSError, IOError):
-            log.warn("Cannot remove, permission denied: {0}".format(path))
-            if trash and move_path_to_trash(path):
-                return
-
-    elif isdir(path):
-        # On Windows, always move to trash first.
-        if trash and on_win and move_path_to_trash(path, preclean=False):
-            return
-
-        try:
-            for i in range(max_retries):
-                try:
-                    shutil.rmtree(path, ignore_errors=False, onerror=warn_failed_remove)
-                    return
-                except OSError as e:
-                    if trash and move_path_to_trash(path):
-                        return
-                    msg = "Unable to delete %s\n%s\n" % (path, e)
-                    if on_win:
-                        try:
-                            shutil.rmtree(path, onerror=_remove_readonly)
-                            return
-                        except OSError as e1:
-                            msg += "Retry with onerror failed (%s)\n" % e1
-
-                        p = subprocess.Popen(['cmd', '/c', 'rd', '/s', '/q', path],
-                                             stdout=subprocess.PIPE,
-                                             stderr=subprocess.PIPE)
-                        (stdout, stderr) = p.communicate()
-                        if p.returncode != 0:
-                            msg += '%s\n%s\n' % (stdout, stderr)
-                        else:
-                            if not isdir(path):
-                                return
-                    log.debug(msg + "Retrying after %s seconds..." % i)
-                    time.sleep(i)
-            # Final time. pass exceptions to caller.
-            shutil.rmtree(path, ignore_errors=False, onerror=warn_failed_remove)
-        finally:
-            # If path was removed, ensure it's not in linked_data_
-            if not isdir(path):
-                delete_linked_data_any(path)
-
-
-def rm_empty_dir(path):
-    """
-    Remove the directory `path` if it is a directory and empty.
-    If the directory does not exist or is not empty, do nothing.
     """
     try:
-        os.rmdir(path)
-    except OSError:  # directory might not exist or not be empty
-        pass
+        with open(path) as fh:
+            for line in fh:
+                line = line.strip()
+                if not line or line.startswith('#'):
+                    continue
+                yield line
+    except (IOError, OSError) as e:
+        if e.errno == errno.ENOENT:
+            raise StopIteration
+        else:
+            raise
 
 
-def yield_lines(path):
-    for line in open(path):
-        line = line.strip()
-        if not line or line.startswith('#'):
-            continue
-        yield line
-
-
-prefix_placeholder = ('/opt/anaconda1anaconda2'
+PREFIX_PLACEHOLDER = ('/opt/anaconda1anaconda2'
                       # this is intentionally split into parts,
                       # such that running this program on itself
                       # will leave it unchanged
                       'anaconda3')
 
+# backwards compatibility for conda-build
+prefix_placeholder = PREFIX_PLACEHOLDER
+
 
 def read_has_prefix(path):
     """
-    reads `has_prefix` file and return dict mapping filenames to
-    tuples(placeholder, mode)
+    reads `has_prefix` file and return dict mapping filepaths to tuples(placeholder, FileMode)
+
+    A line in `has_prefix` contains one of
+      * filepath
+      * placeholder mode filepath
+
+    mode values are one of
+      * text
+      * binary
     """
-    res = {}
-    try:
-        for line in yield_lines(path):
-            try:
-                placeholder, mode, f = [x.strip('"\'') for x in
-                                        shlex.split(line, posix=False)]
-                res[f] = (placeholder, mode)
-            except ValueError:
-                res[line] = (prefix_placeholder, 'text')
-    except IOError:
-        pass
-    return res
+    ParseResult = namedtuple('ParseResult', ('placeholder', 'filemode', 'filepath'))
+
+    def parse_line(line):
+        # placeholder, filemode, filepath
+        parts = tuple(x.strip('"\'') for x in shlex.split(line, posix=False))
+        if len(parts) == 1:
+            return ParseResult(PREFIX_PLACEHOLDER, FileMode.text, parts[0])
+        elif len(parts) == 3:
+            return ParseResult(parts[0], FileMode(parts[1]), parts[2])
+        else:
+            raise RuntimeError("Invalid has_prefix file at path: %s" % path)
+    parsed_lines = (parse_line(line) for line in yield_lines(path))
+    return {pr.filepath: (pr.placeholder, pr.filemode) for pr in parsed_lines}
 
 
-class PaddingError(Exception):
+class _PaddingError(Exception):
     pass
 
 
@@ -369,51 +258,109 @@ def binary_replace(data, a, b):
     replaced with `b` and the remaining string is padded with null characters.
     All input arguments are expected to be bytes objects.
     """
+    if on_win:
+        if has_pyzzer_entry_point(data):
+            return replace_pyzzer_entry_point_shebang(data, a, b)
+        # currently we should skip replacement on Windows for things we don't understand.
+        else:
+            return data
 
     def replace(match):
         occurances = match.group().count(a)
         padding = (len(a) - len(b))*occurances
         if padding < 0:
-            raise PaddingError(a, b, padding)
+            raise _PaddingError
         return match.group().replace(a, b) + b'\0' * padding
+
+    original_data_len = len(data)
     pat = re.compile(re.escape(a) + b'([^\0]*?)\0')
-    res = pat.sub(replace, data)
-    assert len(res) == len(data)
-    return res
+    data = pat.sub(replace, data)
+    assert len(data) == original_data_len
+
+    return data
 
 
 def replace_long_shebang(mode, data):
-    if mode == 'text':
-        shebang_match = re.match(br'^(#!((?:\\ |[^ \n\r])+)(.*))', data)
+    if mode == FileMode.text:
+        shebang_match = SHEBANG_REGEX.match(data)
         if shebang_match:
             whole_shebang, executable, options = shebang_match.groups()
             if len(whole_shebang) > 127:
-                executable_name = executable.decode('utf-8').split('/')[-1]
-                new_shebang = '#!/usr/bin/env {0}{1}'.format(executable_name,
-                                                             options.decode('utf-8'))
-                data = data.replace(whole_shebang, new_shebang.encode('utf-8'))
+                executable_name = executable.decode(UTF8).split('/')[-1]
+                new_shebang = '#!/usr/bin/env %s%s' % (executable_name, options.decode(UTF8))
+                data = data.replace(whole_shebang, new_shebang.encode(UTF8))
     else:
-        pass  # TODO: binary shebangs exist; figure this out in the future if text works well
+        # TODO: binary shebangs exist; figure this out in the future if text works well
+        log.debug("TODO: binary shebangs exist; figure this out in the future if text works well")
     return data
 
 
+def has_pyzzer_entry_point(data):
+    pos = data.rfind(b'PK\x05\x06')
+    return pos >= 0
+
+
+def replace_pyzzer_entry_point_shebang(all_data, placeholder, new_prefix):
+    """Code adapted from pyzzer.  This is meant to deal with entry point exe's created by distlib,
+    which consist of a launcher, then a shebang, then a zip archive of the entry point code to run.
+    We need to change the shebang.
+    https://bitbucket.org/vinay.sajip/pyzzer/src/5d5740cb04308f067d5844a56fbe91e7a27efccc/pyzzer/__init__.py?at=default&fileviewer=file-view-default#__init__.py-112  # NOQA
+    """
+    # Copyright (c) 2013 Vinay Sajip.
+    #
+    # Permission is hereby granted, free of charge, to any person obtaining a copy
+    # of this software and associated documentation files (the "Software"), to deal
+    # in the Software without restriction, including without limitation the rights
+    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+    # copies of the Software, and to permit persons to whom the Software is
+    # furnished to do so, subject to the following conditions:
+    #
+    # The above copyright notice and this permission notice shall be included in
+    # all copies or substantial portions of the Software.
+    #
+    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+    # THE SOFTWARE.
+    launcher = shebang = None
+    pos = all_data.rfind(b'PK\x05\x06')
+    if pos >= 0:
+        end_cdr = all_data[pos + 12:pos + 20]
+        cdr_size, cdr_offset = struct.unpack('<LL', end_cdr)
+        arc_pos = pos - cdr_size - cdr_offset
+        data = all_data[arc_pos:]
+        if arc_pos > 0:
+            pos = all_data.rfind(b'#!', 0, arc_pos)
+            if pos >= 0:
+                shebang = all_data[pos:arc_pos]
+                if pos > 0:
+                    launcher = all_data[:pos]
+
+        if data and shebang and launcher:
+            if hasattr(placeholder, 'encode'):
+                placeholder = placeholder.encode('utf-8')
+            if hasattr(new_prefix, 'encode'):
+                new_prefix = new_prefix.encode('utf-8')
+            shebang = shebang.replace(placeholder, new_prefix)
+            all_data = b"".join([launcher, shebang, data])
+    return all_data
+
+
 def replace_prefix(mode, data, placeholder, new_prefix):
-    if mode == 'text':
-        data = data.replace(placeholder.encode('utf-8'), new_prefix.encode('utf-8'))
-    # Skip binary replacement in Windows.  Some files do have prefix information embedded, but
-    #    this should not matter, as it is not used for things like RPATH.
-    elif mode == 'binary':
-        if not on_win:
-            data = binary_replace(data, placeholder.encode('utf-8'), new_prefix.encode('utf-8'))
-        else:
-            logging.debug("Skipping prefix replacement in binary on Windows")
+    if mode == FileMode.text:
+        data = data.replace(placeholder.encode(UTF8), new_prefix.encode(UTF8))
+    elif mode == FileMode.binary:
+        data = binary_replace(data, placeholder.encode(UTF8), new_prefix.encode(UTF8))
     else:
-        sys.exit("Invalid mode: %s" % mode)
+        raise RuntimeError("Invalid mode: %r" % mode)
     return data
 
 
-def update_prefix(path, new_prefix, placeholder=prefix_placeholder, mode='text'):
-    if on_win:
+def update_prefix(path, new_prefix, placeholder=PREFIX_PLACEHOLDER, mode=FileMode.text):
+    if on_win and mode == FileMode.text:
         # force all prefix replacements to forward slashes to simplify need to escape backslashes
         # replace with unix-style path separators
         new_prefix = new_prefix.replace('\\', '/')
@@ -429,8 +376,6 @@ def update_prefix(path, new_prefix, placeholder=prefix_placeholder, mode='text')
     if data == original_data:
         return
     st = os.lstat(path)
-    # Remove file before rewriting to avoid destroying hard-linked cache
-    os.remove(path)
     with exp_backoff_fn(open, path, 'wb') as fo:
         fo.write(data)
     os.chmod(path, stat.S_IMODE(st.st_mode))
@@ -449,7 +394,7 @@ def dist2pair(dist):
 def dist2quad(dist):
     channel, dist = dist2pair(dist)
     parts = dist.rsplit('-', 2) + ['', '']
-    return (parts[0], parts[1], parts[2], channel)
+    return (str(parts[0]), str(parts[1]), str(parts[2]), str(channel))
 
 
 def dist2name(dist):
@@ -540,12 +485,17 @@ def run_script(prefix, dist, action='post-link', env_prefix=None):
     else:
         shell_path = '/bin/sh' if 'bsd' in sys.platform else '/bin/bash'
         args = [shell_path, path]
-    env = os.environ
-    env['ROOT_PREFIX'] = sys.prefix
-    env['PREFIX'] = str(env_prefix or prefix)
-    env['PKG_NAME'], env['PKG_VERSION'], env['PKG_BUILDNUM'], _ = dist2quad(dist)
+    env = os.environ.copy()
+    env[str('ROOT_PREFIX')] = sys.prefix
+    env[str('PREFIX')] = str(env_prefix or prefix)
+    env[str('PKG_NAME')], env[str('PKG_VERSION')], env[str('PKG_BUILDNUM')], _ = dist2quad(dist)
     if action == 'pre-link':
-        env['SOURCE_DIR'] = str(prefix)
+        sys.stderr.write("""
+Package %s uses a pre-link script. Pre-link scripts are potentially dangerous  discouraged.
+This is because pre-link scripts have the ability to change the package contents in the
+package cache, and therefore modify the underlying files for already-created conda
+environments.  Future versions of conda may deprecate and ignore pre-link scripts.\n""" % dist)
+        env[str('SOURCE_DIR')] = str(prefix)
     try:
         subprocess.check_call(args, env=env)
     except subprocess.CalledProcessError:
@@ -563,20 +513,15 @@ def read_icondata(source_dir):
 
     try:
         data = open(join(source_dir, 'info', 'icon.png'), 'rb').read()
-        return base64.b64encode(data).decode('utf-8')
+        return base64.b64encode(data).decode(UTF8)
     except IOError:
         pass
     return None
 
 
 def read_no_link(info_dir):
-    res = set()
-    for fn in 'no_link', 'no_softlink':
-        try:
-            res.update(set(yield_lines(join(info_dir, fn))))
-        except IOError:
-            pass
-    return res
+    return set(chain(yield_lines(join(info_dir, 'no_link')),
+                     yield_lines(join(info_dir, 'no_softlink'))))
 
 
 # Should this be an API function?
@@ -607,7 +552,7 @@ def symlink_conda_hlp(prefix, root_dir, where, symlink_fn):
         try:
             # try to kill stale links if they exist
             if os.path.lexists(prefix_file):
-                os.remove(prefix_file)
+                rm_rf(prefix_file)
             # if they're in use, they won't be killed.  Skip making new symlink.
             if not os.path.lexists(prefix_file):
                 symlink_fn(root_file, prefix_file)
@@ -619,6 +564,7 @@ def symlink_conda_hlp(prefix, root_dir, where, symlink_fn):
             else:
                 raise
 
+
 # ========================== begin API functions =========================
 
 def try_hard_link(pkgs_dir, prefix, dist):
@@ -641,7 +587,7 @@ def try_hard_link(pkgs_dir, prefix, dist):
         return False
     finally:
         rm_rf(dst)
-        rm_empty_dir(prefix)
+
 
 # ------- package cache ----- construction
 
@@ -688,11 +634,11 @@ def add_cached_package(pdir, url, overwrite=False, urlstxt=False):
     if not (xpkg or xdir):
         return
     if url:
-        url = remove_binstar_tokens(url)
-    _, schannel = url_channel(url)
+        url = url
+    schannel = Channel(url).canonical_name
     prefix = '' if schannel == 'defaults' else schannel + '::'
     xkey = xpkg or (xdir + '.tar.bz2')
-    fname_table_[xkey] = fname_table_[url_path(xkey)] = prefix
+    fname_table_[xkey] = fname_table_[path_to_url(xkey)] = prefix
     fkey = prefix + dist
     rec = package_cache_.get(fkey)
     if rec is None:
@@ -725,7 +671,8 @@ def package_cache():
         return package_cache_
     # Stops recursion
     package_cache_['@'] = None
-    for pdir in pkgs_dirs:
+    # import pdb; pdb.set_trace()
+    for pdir in context.pkgs_dirs:
         try:
             data = open(join(pdir, 'urls.txt')).read()
             for url in data.split()[::-1]:
@@ -762,7 +709,7 @@ def find_new_location(dist):
     # Look for a location with no conflicts
     # On the second pass, just pick the first location
     for p in range(2):
-        for pkg_dir in pkgs_dirs:
+        for pkg_dir in context.pkgs_dirs:
             pkg_path = join(pkg_dir, fname)
             prefix = fname_table_.get(pkg_path)
             if p or prefix is None:
@@ -796,12 +743,16 @@ def rm_fetched(dist):
         return
     for fname in rec['files']:
         del fname_table_[fname]
-        del fname_table_[url_path(fname)]
-        with Locked(dirname(fname)):
+        del fname_table_[path_to_url(fname)]
+        with FileLock(fname):
             rm_rf(fname)
+            if exists(fname):
+                log.warn("File not removed during RM_FETCHED instruction: %s", fname)
     for fname in rec['dirs']:
-        with Locked(dirname(fname)):
+        with FileLock(fname):
             rm_rf(fname)
+            if exists(fname):
+                log.warn("Directory not removed during RM_FETCHED instruction: %s", fname)
     del package_cache_[dist]
 
 
@@ -831,8 +782,10 @@ def rm_extracted(dist):
     if rec is None:
         return
     for fname in rec['dirs']:
-        with Locked(dirname(fname)):
+        with FileLock(fname):
             rm_rf(fname)
+            if exists(fname):
+                log.warn("Directory not removed during RM_EXTRACTED instruction: %s", fname)
     if rec['files']:
         rec['dirs'] = []
     else:
@@ -849,8 +802,8 @@ def extract(dist):
     fname = rec['files'][0]
     assert url and fname
     pkgs_dir = dirname(fname)
-    with Locked(pkgs_dir):
-        path = fname[:-8]
+    path = fname[:-8]
+    with FileLock(path):
         temp_path = path + '.tmp'
         rm_rf(temp_path)
         with tarfile.open(fname) as t:
@@ -867,17 +820,17 @@ def extract(dist):
                     os.lchown(p, 0, 0)
         add_cached_package(pkgs_dir, url, overwrite=True)
 
+
 # Because the conda-meta .json files do not include channel names in
 # their filenames, we have to pull that information from the .json
 # files themselves. This has made it necessary in virtually all
 # circumstances to load the full set of files from this directory.
 # Therefore, we have implemented a full internal cache of this
 # data to eliminate redundant file reads.
-
 linked_data_ = {}
 
 
-def load_linked_data(prefix, dist, rec=None):
+def load_linked_data(prefix, dist, rec=None, ignore_channels=False):
     schannel, dname = dist2pair(dist)
     meta_file = join(prefix, 'conda-meta', dname + '.json')
     if rec is None:
@@ -900,13 +853,16 @@ def load_linked_data(prefix, dist, rec=None):
         channel = channel.rstrip('/')
         if not url or (url.startswith('file:') and channel[0] != '<unknown>'):
             url = rec['url'] = channel + '/' + fn
-    channel, schannel = url_channel(url)
+    channel, schannel = Channel(url).url_channel_wtf
     rec['url'] = url
     rec['channel'] = channel
     rec['schannel'] = schannel
     rec['link'] = rec.get('link') or True
-    cprefix = '' if schannel == 'defaults' else schannel + '::'
-    linked_data_[prefix][str(cprefix + dname)] = rec
+    if ignore_channels:
+        linked_data_[prefix][dname] = rec
+    else:
+        cprefix = '' if schannel == 'defaults' else schannel + '::'
+        linked_data_[prefix][str(cprefix + dname)] = rec
     return rec
 
 
@@ -917,7 +873,7 @@ def delete_linked_data(prefix, dist, delete=True):
     if delete:
         meta_path = join(prefix, 'conda-meta', dist2filename(dist, '.json'))
         if isfile(meta_path):
-            os.unlink(meta_path)
+            rm_rf(meta_path)
 
 
 def delete_linked_data_any(path):
@@ -944,7 +900,7 @@ def load_meta(prefix, dist):
     return linked_data(prefix).get(dist)
 
 
-def linked_data(prefix):
+def linked_data(prefix, ignore_channels=False):
     """
     Return a dictionary of the linked packages in prefix.
     """
@@ -956,14 +912,15 @@ def linked_data(prefix):
         if isdir(meta_dir):
             for fn in os.listdir(meta_dir):
                 if fn.endswith('.json'):
-                    load_linked_data(prefix, fn[:-5])
+                    load_linked_data(prefix, fn[:-5], ignore_channels=ignore_channels)
     return recs
 
-def linked(prefix):
+
+def linked(prefix, ignore_channels=False):
     """
     Return the set of canonical names of linked packages in prefix.
     """
-    return set(linked_data(prefix).keys())
+    return set(linked_data(prefix, ignore_channels=ignore_channels).keys())
 
 
 def is_linked(prefix, dist):
@@ -975,108 +932,68 @@ def is_linked(prefix, dist):
     return load_meta(prefix, dist)
 
 
-def delete_trash(prefix=None):
-    for pkg_dir in pkgs_dirs:
-        trash_dir = join(pkg_dir, '.trash')
-        if not isdir(trash_dir):
-            continue
-        try:
-            log.debug("Trying to delete the trash dir %s" % trash_dir)
-            rm_rf(trash_dir, max_retries=1, trash=False)
-        except OSError as e:
-            log.debug("Could not delete the trash dir %s (%s)" % (trash_dir, e))
-
-
-def move_to_trash(prefix, f, tempdir=None):
-    """
-    Move a file or folder f from prefix to the trash
-
-    tempdir is a deprecated parameter, and will be ignored.
-
-    This function is deprecated in favor of `move_path_to_trash`.
-    """
-    return move_path_to_trash(join(prefix, f) if f else prefix)
-
-
-def move_path_to_trash(path, preclean=True):
-    """
-    Move a path to the trash
-    """
-    # Try deleting the trash every time we use it.
-    if preclean:
-        delete_trash()
-
-    for pkg_dir in pkgs_dirs:
-        trash_dir = join(pkg_dir, '.trash')
-
-        try:
-            os.makedirs(trash_dir)
-        except OSError as e1:
-            if e1.errno != errno.EEXIST:
-                continue
-
-        trash_file = tempfile.mktemp(dir=trash_dir)
-
-        try:
-            os.rename(path, trash_file)
-        except OSError as e:
-            log.debug("Could not move %s to %s (%s)" % (path, trash_file, e))
-        else:
-            log.debug("Moved to trash: %s" % (path,))
-            delete_linked_data_any(path)
-            if not preclean:
-                rm_rf(trash_file, max_retries=1, trash=False)
-            return True
-
-    return False
-
-
 def link(prefix, dist, linktype=LINK_HARD, index=None):
     """
     Set up a package in a specified (environment) prefix.  We assume that
     the package has been extracted (using extract() above).
     """
+    log.debug("linking package %s with link type %s", dist, linktype)
     index = index or {}
     source_dir = is_extracted(dist)
     assert source_dir is not None
     pkgs_dir = dirname(source_dir)
-    log.debug('pkgs_dir=%r, prefix=%r, dist=%r, linktype=%r' %
-              (pkgs_dir, prefix, dist, linktype))
+    log.debug('pkgs_dir=%r, prefix=%r, dist=%r, linktype=%r', pkgs_dir, prefix, dist, linktype)
 
     if not run_script(source_dir, dist, 'pre-link', prefix):
-        sys.exit('Error: pre-link failed: %s' % dist)
+        raise LinkError('Error: pre-link failed: %s' % dist)
 
     info_dir = join(source_dir, 'info')
+
+    if not os.path.isfile(join(info_dir, "files")):
+        print("Installing %s requires a minimum conda version of 4.3." % dist, file=sys.stderr)
+        sys.exit(1)
+
     files = list(yield_lines(join(info_dir, 'files')))
     has_prefix_files = read_has_prefix(join(info_dir, 'has_prefix'))
     no_link = read_no_link(info_dir)
 
-    with Locked(prefix), Locked(pkgs_dir):
-        for f in files:
-            src = join(source_dir, f)
-            dst = join(prefix, f)
+    full_dist_name = "%s.tar.bz2" % dist
+    if not index.get(full_dist_name, {}).get("noarch", None) in (True, False, None, "generic"):
+        print("Installing %s requires a minimum conda version of 4.3." % dist, file=sys.stderr)
+        sys.exit(1)
+
+    # for the lock issue
+    # may run into lock if prefix not exist
+    if not isdir(prefix):
+        os.makedirs(prefix)
+
+    with DirectoryLock(prefix), FileLock(source_dir):
+        for filepath in files:
+            src = join(source_dir, filepath)
+            dst = join(prefix, filepath)
+
             dst_dir = dirname(dst)
             if not isdir(dst_dir):
                 os.makedirs(dst_dir)
             if os.path.exists(dst):
-                log.warn("file already exists: %r" % dst)
+                log.info("file exists, but clobbering: %r" % dst)
                 rm_rf(dst)
             lt = linktype
-            if f in has_prefix_files or f in no_link or islink(src):
+            if filepath in has_prefix_files or filepath in no_link or islink(src):
                 lt = LINK_COPY
+
             try:
                 _link(src, dst, lt)
             except OSError as e:
-                log.error('failed to link (src=%r, dst=%r, type=%r, error=%r)' %
-                          (src, dst, lt, e))
+                raise CondaOSError('failed to link (src=%r, dst=%r, type=%r, error=%r)' %
+                                   (src, dst, lt, e))
 
-        for f in sorted(has_prefix_files):
-            placeholder, mode = has_prefix_files[f]
+        for filepath in sorted(has_prefix_files):
+            placeholder, mode = has_prefix_files[filepath]
             try:
-                update_prefix(join(prefix, f), prefix, placeholder, mode)
-            except PaddingError:
-                sys.exit("ERROR: placeholder '%s' too short in: %s\n" %
-                         (placeholder, dist))
+                update_prefix(join(prefix, filepath), prefix, placeholder, mode)
+            except _PaddingError:
+                raise PaddingError(dist, placeholder, len(placeholder))
 
         # make sure that the child environment behaves like the parent,
         #    wrt user/system install on win
@@ -1086,19 +1003,19 @@ def link(prefix, dist, linktype=LINK_HARD, index=None):
             if isfile(nonadmin):
                 open(join(prefix, ".nonadmin"), 'w').close()
 
-        if config.shortcuts:
+        if context.shortcuts:
             mk_menus(prefix, files, remove=False)
 
         if not run_script(prefix, dist, 'post-link'):
-            sys.exit("Error: post-link failed for: %s" % dist)
+            raise LinkError("Error: post-link failed for: %s" % dist)
 
         meta_dict = index.get(dist + '.tar.bz2', {})
         meta_dict['url'] = read_url(dist)
-        try:
-            alt_files_path = join(prefix, 'conda-meta', dist2filename(dist, '.files'))
+        alt_files_path = join(prefix, 'conda-meta', dist2filename(dist, '.files'))
+        if isfile(alt_files_path):
+            # alt_files_path is a hack for noarch
             meta_dict['files'] = list(yield_lines(alt_files_path))
-            os.unlink(alt_files_path)
-        except IOError:
+        else:
             meta_dict['files'] = files
         meta_dict['link'] = {'source': source_dir,
                              'type': link_name_map.get(linktype)}
@@ -1113,7 +1030,8 @@ def unlink(prefix, dist):
     Remove a package from the specified environment, it is an error if the
     package does not exist in the prefix.
     """
-    with Locked(prefix):
+    with DirectoryLock(prefix):
+        log.debug("unlinking package %s", dist)
         run_script(prefix, dist, 'pre-unlink')
 
         meta = load_meta(prefix, dist)
@@ -1138,115 +1056,19 @@ def unlink(prefix, dist):
         dst_dirs2.add(join(prefix, 'conda-meta'))
         dst_dirs2.add(prefix)
 
+        # remove empty directories
         for path in sorted(dst_dirs2, key=len, reverse=True):
-            rm_empty_dir(path)
+            if isdir(path) and not os.listdir(path):
+                rm_rf(path)
 
 
 def messages(prefix):
     path = join(prefix, '.messages.txt')
     try:
         with open(path) as fi:
-            sys.stdout.write(fi.read())
+            fh = sys.stderr if context.json else sys.stdout
+            fh.write(fi.read())
     except IOError:
         pass
     finally:
         rm_rf(path)
-
-
-def duplicates_to_remove(dist_metas, keep_dists):
-    """
-    Returns the (sorted) list of distributions to be removed, such that
-    only one distribution (for each name) remains.  `keep_dists` is an
-    iterable of distributions (which are not allowed to be removed).
-    """
-    from collections import defaultdict
-
-    keep_dists = set(keep_dists)
-    ldists = defaultdict(set)  # map names to set of distributions
-    for dist in dist_metas:
-        name = name_dist(dist)
-        ldists[name].add(dist)
-
-    res = set()
-    for dists in ldists.values():
-        # `dists` is the group of packages with the same name
-        if len(dists) == 1:
-            # if there is only one package, nothing has to be removed
-            continue
-        if dists & keep_dists:
-            # if the group has packages which are have to be kept, we just
-            # take the set of packages which are in group but not in the
-            # ones which have to be kept
-            res.update(dists - keep_dists)
-        else:
-            # otherwise, we take lowest (n-1) (sorted) packages
-            res.update(sorted(dists)[:-1])
-    return sorted(res)
-
-
-# =========================== end API functions ==========================
-
-
-def main():
-    # This CLI is only invoked from the self-extracting shell installers
-    global pkgs_dirs
-    from optparse import OptionParser
-
-    p = OptionParser(description="conda link tool used by installer")
-
-    p.add_option('--file',
-                 action="store",
-                 help="path of a file containing distributions to link, "
-                      "by default all packages extracted in the cache are "
-                      "linked")
-
-    p.add_option('--prefix',
-                 action="store",
-                 default=sys.prefix,
-                 help="prefix (defaults to %default)")
-
-    p.add_option('-v', '--verbose',
-                 action="store_true")
-
-    opts, args = p.parse_args()
-    if args:
-        p.error('no arguments expected')
-
-    logging.basicConfig()
-
-    prefix = opts.prefix
-    pkgs_dir = join(prefix, 'pkgs')
-    pkgs_dirs = [pkgs_dir]
-    if opts.verbose:
-        print("prefix: %r" % prefix)
-
-    if opts.file:
-        idists = list(yield_lines(join(prefix, opts.file)))
-    else:
-        idists = sorted(extracted())
-    assert idists
-
-    linktype = (LINK_HARD
-                if try_hard_link(pkgs_dir, prefix, idists[0]) else
-                LINK_COPY)
-    if opts.verbose:
-        print("linktype: %s" % link_name_map[linktype])
-
-    for dist in idists:
-        if opts.verbose:
-            print("linking: %s" % dist)
-        link(prefix, dist, linktype)
-
-    messages(prefix)
-
-    for dist in duplicates_to_remove(linked(prefix), idists):
-        meta_path = join(prefix, 'conda-meta', dist + '.json')
-        print("WARNING: unlinking: %s" % meta_path)
-        try:
-            os.rename(meta_path, meta_path + '.bak')
-        except OSError:
-            rm_rf(meta_path)
-
-
-if __name__ == '__main__':
-    main()
